# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
        Classes for the simulation of neutral and charged particles in a plasma discharge.

        This module contains a class for managing the neutral particles (mainly atoms and molecules) 
        which are the targets of collisions by moving particles, usually electrons and ions.
        These target particles are considered fixed in space, since their speed is much lower respect to moving particles.
        Cross sections needed to treat the collision processes are imported from ASCII files.

        How collisions are treated
        --------------------------

        To decide if an electron (or ion) scatters whith a neutral (by any type of scattering process: elastic, 
        ionization, dissociation of any type) during a time interval dt, we consider that the 
        probability for (at least) a scattering event to happen during a time interval dt is
        p = 1 - exp( f(v) * dt ) = 1 - exp( r(v) * v * dt ) = 1 - exp( s(v) * n * v * dt )
        where: 
                v           -> electron speed
                n           -> total neutral density (sum of densities of all neutral types)
                f(v)=r(v)*v -> scattering frequency (number of scattering events experienced by a single electron in a unit time)
                r(v)=s(v)*n -> total scattering rate (number of scattering events for a unit length of 
                               path travelled by the electron)
                s(v)        -> total cross section (sum of all cross sections for all scattering types) 
                               at speed v

        Once decided that a scattering has happened, to decide of which type it was, we have to use
        relative probabilities and probability limits.
        The relative probability p_i that the scattering was of type i is
                p_i = r_i(v)/r(v)
        where:
                i      -> identifies the type of scattering process
                          0            <= i <= self.types-1   -> elastic scattering (with neutral type i)
                          self.types   <= i <= 2*self.types-1 -> ionization 
                                                                        (of neutral type i-self.types)
                          2*self.types <= i                   -> dissociation 
                                                                        (neutral type and dissociation
                                                                         process depending on value of i)
                r_i(v) -> scattering rate for process i at speed v
                r(v)   -> total scattering rate for all processes at speed v
                of course, the sum of all p_i gives 1.
                To decide the type of scattering, we extract a random numbed rand between 0 and 1, than compare the
                        number with the limits PL_i defined as follows
                        PL_0 = p_0
                        PL_i = PL_(i-1) + p_i           for i>0
                The type of scattering is given by the value of i for which it is true that
                        PL_(i-1) < rand < PL_i          whith i>0
                        or
                        0        < rand < PL_0          if i=0

        Recombination
        -------------

        The same considerations hold for recombination of electrons with ions, by the three main recombination channels:
        dissociative recombination, three body recombination and radiative recombination.
        In this case, we must substuitute the neutral density n with the ion density n_ion. However, while the former
        can be considered approximately constant if the ionization degree is very low, the latter is changing in time
        as the ionization process proceeds.

        Documentation is also available in the docstrings.
"""



# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

import math
import numpy

from pysica.parameters import *
from pysica.constants import *
from pysica.plasmapro.ccpla_defaults import *     
from pysica.functions.physics import number_density
from pysica.managers.data_manager import DataGrid
from pysica.managers.gnuplot_manager import gnuplot_installed, new_plot, plot2d, plot_curves, plot_close, plot_close_all
from pysica.io.io_files import write_to_disk
from pysica.plasmapro.discharge.particles_data_manager import read_file_ntypes, read_file_properties


XSEC_PLOT_PERSIST = True

# +------------------+
# | Target particles |
# +------------------+

class TargetParticles:
        """A class defining a collection of particles, to be considered as the targets of collisions by electrons and ions."""

        def __init__(self, N_sigma_electrons, N_sigma_ions, temperature, total_pressure, min_scattered=DEFAULT_MIN_SCATTERED,
                     isactive_recomb=True, neutral_types=None, filename=None):
                """ Initialises the properties of target particles.

                        Initialises the properties of target particles.
                        If the number of target particle types is not specified, it will be read from a ASCII file containing
                        particles properties.

                        Parameters
                        ----------

                        N_sigma_electrons: number of energies for which cross section for electrons scattering values are tabulated
                        N_sigma_ions:      number of energies for which cross section for ions scatterig values are tabulated
                        temperature:       temperature of gases in K
                        total_pressure:    total pressure of gases in Pa
                        isactive_recomb:   set if electron/ion recombination processes must be activated or not
                        neutral_types:     number of neutral types, if this value is equal to None
                                              than the number of neutrals will be read from 
                                              the ASCII file with name filename
                        filename:          name of the ASCII file containing neutrals properties
                                              if neutral_types=None this parameter cannot be left blank

                        Initialized data attributes
                        ---------------------------

                        self.read_error:                used for error checking
                                                        (status, message)
                                                                status: 0 = no error
                                                                        1 = file not found
                                                                        2 = no gases were given in the file
                                                                message: an error message or 'OK'
                        self.temperature:               temperature of the gas mixture / K
                        self.total_pressure:            total pressure of the gas mixture / Pa
                        self.isactive_recomb:           set if electron/ion recombination processes must be activated or not
                        self.types:                     how many types of target molcules are involved
                        self.n_sigma:                   number of velocity values for which electron impact cross sections 
                                                               are tabulated
                        self.n_sigma_ions:              number of velocity values for which ion impact cross sections are tabulated
                """

                self.read_error = (0, 'OK')

                # If the number of gases was not given, try to read it from file
                if (neutral_types is None):
                        if (filename is None):
                                self.read_error = (3, "Filename was not given")
                                return
                        else:
                                self.read_error = read_file_ntypes(filename, self)
                                if (self.read_error[0] != 0): return
                else:
                        self.types              = neutral_types

                # Characteristics of the plasma discharge
                self.temperature                = temperature
                self.total_pressure             = total_pressure
                self.isactive_recomb            = isactive_recomb

                # Number of velocity values used to discretize cross sections for electron and ion impact with neutrals
                self.n_sigma                    = N_sigma_electrons
                self.n_sigma_ions               = N_sigma_ions
                self.min_scattered              = min_scattered


        # +-------------------------------------------------------+
        # | Get energies for which cross section values are known |
        # +-------------------------------------------------------+

        def sigma_energies(self):
                """Returns an array with the energies (in eV) for which cross section values for electrons are tabulated.

                        Returns
                        -------

                        array of the energies (expressed in eV) for which the cross section values are tabulated

                """

                return - 0.5 * ELECTRON_MASS / ELECTRON_CHARGE * self.sigma_velocities**2


        def sigma_energies_ions(self, ion_type):
                """Returns an array with the energies (in eV) for which cross section values for ions are tabulated.

                        Parameters
                        ----------

                        ion_type: the type of ion for the energies are requested


                        Returns
                        -------

                        array of the energies (expressed in eV) for which the cross section values are tabulated

                """

                return - 0.5 * self.mass[ion_type] * ATOMIC_UNIT_MASS / ELECTRON_CHARGE * self.sigma_velocities_ions[ion_type]**2


        # +--------------------------------+
        # | Get the properties of neutrals |
        # +--------------------------------+

        def read_properties(self, filename, sep, e_min_sigma, e_max_sigma, e_min_sigma_ions, e_max_sigma_ions, debug=False):
                """Reads the neutrals characteristics from an ascii file and calculates some properties.

                        Parameters
                        ----------

                        filename:         name of file containing neutral gases characteristics
                        sep:              character used in the ASCII file to separate values (usually the tab char)
                        e_min_sigma:      minimum energy for which cross sections for electron scattering are tabulated
                        e_max_sigma:      maximum energy for which cross sections for electron scattering are tabulated
                        e_min_sigma_ions: minimum energy for which cross sections for ion scattering are tabulated
                        e_max_sigma_ions: maximum energy for which cross sections for ion scattering are tabulated

                        Initialized data attributes
                        ---------------------------
                        self.names:                      names of the gas molecules (max 8 char)
                        self.molecule_type:              string describing the type of molecule (1 char)
                                                                'a'= atom, 'm'= nonpolar molecule, 'p'= polar molecule
                        self.mass:                       masses of gas molecules / u
                        self.mean_v:                     mean speeds of gas molecules (all at the same temperature)
                        self.gas_flow:                   array of gas fluxes / arbitrary units
                        self.partial_pressure:           array of gases partial pressures / Pa
                        self.number_density:             number density of neutral species
                        self.secondary_emission:         secondary emission coeff. i.e. average number of e- emitted 
                                                                when an ion is collected at the cathode
                        self.dissociation_rate           dissociation rate for each molecule type and its dissociation types
                        self.dissociation_rate_constant  dissociation rate constant for each molecule type and its dissociation types

                        self.energy_loss:                fraction of energy transferred from electron to neutral 
                                                                during elastic scattering
                        self.energy_loss_ions:           fraction of energy transferred from ions to neutral 
                                                                during elastic scattering
                        self.v_ratio:                    electron speed reduction during elastic scattering
                        self.v_ratio_ions:               ions speed reduction during elastic scattering
                        self.ionization_energy:          first ionization energy / eV
                        self.dissociation_types:         number of dissociation processes for each neutral type
                        self.dissociation_energy:        dissociation energies for processes to be considered / eV

                        self.sigma_velocities:           array of velocity values for which electron impact cross sections 
                                                                are tabulated
                        self.sigma_velocities_ions:      matrix of velocity values for which cross sections are tabulated 
                                                                for each ion type

                        self.sigma_elastic:              cross section table for electron impact elastic scattering 
                        self.sigma_ionization:           cross section table for electron impact ionization
                        self.sigma_dissociation:         cross section table for electron impact dissociation   

                        self.sigma_recombination_diss    cross section table for electron impact dissociative recombination 
                                                                of ionized molecules
#                       self.sigma_recombination_3body   cross section table for electron impact three body recombination 
                                                                of ionized atoms/molecules
#                       self.sigma_recombination_rad     cross section table for electron impact radiative recombination 
                                                                of ionized atoms/molecules

                        self.sigma_elastic_ions:         cross section table for ions impact elastic scattering 
                        self.sigma_charge_exchange_ions: cross section table for ions impact elastic scattering 

                        self.collisions_null:             number of electron null collisions
                        self.collisions_elastic:          number of electron elastic collisions
                        self.collisions_ionization:       number of electron ionization collisions
                        self.collisions_dissociation:     number of electron dissociation collisions
                        self.collisions_recombination:    number of electron/ion recombination collisions
                        self.collisions_total_electron    total number of electron collisions, of any type
                        self.dissociation_rate:           dissociation rate of neutral particles, 
                                                                for each type and dissociation process
                        self.dissociation_rate_constant:  dissociation rate constant of neutral particles, 
                                                                for each type and dissociation process

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = could not open file
                                         2 = missing values
                                         3 = numerical value expected
                                         4 = unknown molecule type
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                # Characteristics of gases
                self.names                      = numpy.zeros(self.types, 'U8')
                self.molecule_type              = numpy.zeros(self.types, 'U1')
                self.mass                       = numpy.zeros(self.types, 'd')
                self.mean_v                     = numpy.zeros(self.types, 'd')
                self.gas_flow                   = numpy.zeros(self.types, 'd')
                self.partial_pressure           = numpy.zeros(self.types, 'd')
                self.number_density             = numpy.zeros(self.types, 'd')
                self.secondary_emission         = numpy.zeros(self.types, 'd')

                # Characteristics for interaction with electrons
                self.energy_loss                = numpy.zeros(self.types, 'd')
                self.v_ratio                    = numpy.zeros(self.types, 'd')
                self.ionization_energy          = numpy.zeros(self.types, 'd')
                self.dissociation_types         = numpy.zeros(self.types, 'b')
                self.dissociation_energy        = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES), 'd', order='F' )

                # Characteristics for interaction with ions
                self.energy_loss_ions           = numpy.zeros( (self.types, self.types), 'd', order='F' )
                self.v_ratio_ions               = numpy.zeros( (self.types, self.types), 'd', order='F' )

                # Arrays of velocity values for which cross sections are tabulated
                self.sigma_velocities           = numpy.zeros(self.n_sigma, 'd')        
                self.sigma_velocities_ions      = numpy.zeros( (self.types, self.n_sigma_ions), 'd', order='F') 

                # Arrays of cross sections for electron impact processes
                self.sigma_elastic              = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')
                self.sigma_ionization           = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')
                self.sigma_dissociation         = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES, self.n_sigma), 
                                                             'd', order = 'F')

                self.sigma_recombination_diss   = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES, self.n_sigma), 
                                                             'd', order = 'F')
                self.sigma_recombination_3body  = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')
                self.sigma_recombination_rad    = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')

                # Arrays of cross sections for ion impact processes
                self.sigma_elastic_ions         = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')
                self.sigma_charge_exchange_ions = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')

################# Quantities that will be calculated during simulation
                # collision counters for electron/neutral collisions and electron/ion recombinations
                # and dissociation rate / rate constant counters
                self.collisions_null            = numpy.zeros(1, 'd') # Must be an array since it will be passed to f2py
                self.collisions_elastic         = numpy.zeros(self.types, 'd')
                self.collisions_ionization      = numpy.zeros(self.types, 'd')
                self.collisions_dissociation    = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES), 'd', order='F' )
                self.collisions_recombination   = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES), 'd', order='F' )
                self.collisions_total_electron  = 0
                self.dissociation_rate          = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES), 'd', order='F' )
                self.dissociation_rate_constant = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES), 'd', order='F' )               
################
               
                # Read neutrals properties from ascii file
                (status, message) = read_file_properties(filename, sep, self, debug)
                if (status != 0): return (status, message)

                # Calculate neutrals properties
                self.partial_pressure   = self.gas_flow * self.total_pressure / self.gas_flow.sum()
                self.number_density     = number_density(self.partial_pressure, self.temperature)
                # Average electron energy loss in elastic scattering
                self.energy_loss        = 2 * ELECTRON_MASS / (self.mass * ATOMIC_UNIT_MASS) 
                self.v_ratio            = numpy.sqrt(numpy.abs(1-self.energy_loss))                     
                self.mean_v             = numpy.sqrt(8 * K_BOLTZMANN * self.temperature / (math.pi * self.mass))

                for i in range(self.types):
                        for j in range(self.types):
                                # Average ion energy loss in elastic scattering
                                self.energy_loss_ions[i][j] = 2 * self.mass[i] * self.mass[j] / (self.mass[i] + self.mass[j])**2
                                self.v_ratio_ions[i][j]     = numpy.sqrt(1-self.energy_loss_ions[i][j])

                # Calculate speed values at which cross sections for electrons will be interpolated
                v_min_sigma = math.sqrt( 2 * e_min_sigma * (-ELECTRON_CHARGE / ELECTRON_MASS) )
                v_max_sigma = math.sqrt( 2 * e_max_sigma * (-ELECTRON_CHARGE / ELECTRON_MASS) )
                self.sigma_velocities = numpy.linspace(v_min_sigma, v_max_sigma, self.n_sigma )

                # Calculate speed values at which cross sections for each ion type will be interpolated
                for i in range(self.types):
                        v_min_sigma = math.sqrt( 2 * e_min_sigma_ions * (-ELECTRON_CHARGE / (self.mass[i] * ATOMIC_UNIT_MASS)) )
                        v_max_sigma = math.sqrt( 2 * e_max_sigma_ions * (-ELECTRON_CHARGE / (self.mass[i] * ATOMIC_UNIT_MASS)) )
                        self.sigma_velocities_ions[i] = numpy.linspace(v_min_sigma, v_max_sigma, self.n_sigma_ions)

                return (status, message)


        # +--------------------------------------------------------------------+
        # | General method used to read cross section values from a ascii file |
        # +--------------------------------------------------------------------+

        def _read_xsec(self, filename, sep, qm_ratio, target_index, ion_index=None, plot=False, title=''):
                """Reads a table of cross section values as a function of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form
                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment
                                        where energies are in eV and cross section values are in m**2
                                        
                        sep:            character used to separate values (usually the tab char)

                        qm_ratio:       the charge to mass ratio of incident particle

                        ion_index:      integer value indentifying the type of ion (i.e. the type of neutral that was ionized) 
                                        that scatters with the target, 
                                        if it is None, than the particle is an electron

                        target_index:   integer value indentifying the type of target

                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation

                        title:          a string containing the title to be printed on the plot


                        Initialized data attributes
                        ---------------------------

                        self.sigma_return       numpy array containing the cross section values that were read from the file


                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                         3 = ion index out of bounds
                                message: a string containing an error message or 'Ok'
                """


                status, message  = 0, 'OK'              # This variable is used for error checking      

                if (ion_index is None):
                        self.sigma_return= numpy.zeros(self.n_sigma, 'd')
                else:
                        if ( (ion_index < 0) or (ion_index > self.types-1)):
                                status = 3
                                message = 'ion type ' + str(ion_index) + ' is out of bounds'
                                return(status, message)                                
                        self.sigma_return= numpy.zeros(self.n_sigma_ions, 'd')

                # Read the tabulated values of elastic cross-section into a numpy array
                sigma = DataGrid()
                (status, message) = sigma.read_file(filename, n_columns=2, sep=sep, transpose=True)
                if (status != 0):
                        return (status, message)
                else: 
                        sigma.n_data = len(sigma.data_array[0])

                # Get the absolute value of mass to charge ratio
                if (qm_ratio < 0): qm_ratio = - qm_ratio

                # Check that all cross-section values are positive
                for j in range(sigma.n_data):
                        if sigma.data_array[0][j] < 0:  
                                status  = 1
                                message = 'negative energy value found in row # '+str(j+1)
                                return (status, message)
                        if sigma.data_array[1][j] < 0:
                                status  = 2
                                message = 'negative cross-section value found in row # '+str(j+1)
                                return (status, message)

                # Change from energies [eV] to velocities [m/s]
                sigma.data_array[0] = numpy.sqrt( 2 * qm_ratio * sigma.data_array[0] )

                # Calculate a interpolated cross-section table
                if (ion_index is None): 
                        self.sigma_return = numpy.interp(self.sigma_velocities,
                                                         sigma.data_array[0], sigma.data_array[1])
                else:
                        self.sigma_return = numpy.interp(self.sigma_velocities_ions[ion_index],
                                                         sigma.data_array[0], sigma.data_array[1])

                # Plot cross-section
                if (plot and gnuplot_installed): 
                        self.xsec_graph = new_plot(title=title,
                                                   xlabel='Energy / eV',
                                                   ylabel='Cross section / m**2',
                                                   format_x='%g',
                                                   format_y='%g',
                                                   logx=True,
                                                   grid=True,
                                                   persistence=XSEC_PLOT_PERSIST,
                                                   redirect_output=True)
                        self.plot_list = []
                        
                        # These are the values given in the data file
                        self.plot_list.append( [0.5 / qm_ratio * sigma.data_array[0]**2,
                                                sigma.data_array[1],
                                                'Tabulated values' ] )
                        
                        # These are the values calculated by interpolation of the ones in the data file
                        if (ion_index is None):
                                self.plot_list.append( [ self.sigma_energies(),
                                                         self.sigma_return,
                                                         'Interpolated values' ] )
                        else:
                                self.plot_list.append( [ self.sigma_energies_ions(ion_index),
                                                         self.sigma_return,
                                                         'Interpolated values' ] )
                        # Close the plots, to quit the associated gnuplot processes
                        # plots remain visible thanks to the "persist" option
                        plot_curves(self.xsec_graph, self.plot_list)
                        plot_close(self.xsec_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.xsec_graph)

                return (status, message)


        # +------------------------------------------------------------------+
        # | Methods to read cross sections for electrons/neutrals scattering |
        # +------------------------------------------------------------------+

        def read_xsec_electrons_elastic(self, filename, sep, neutral_index, plot=False):
                """Reads the table of elastic cross section values as a function of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form

                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                        
                        sep:            character used to separate values (usually the tab char)
                        neutral_index:  integer value indentifying the type of neutral for which cross section
                                        shuld be read

                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation


                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                title    = 'Cross section for elastic scattering of e- on '+self.names[neutral_index]
                qm_ratio = - ELECTRON_CHARGE / ELECTRON_MASS

                (status, message) = self._read_xsec(filename, sep, qm_ratio, neutral_index, None, plot, title)

                if (status == 0): self.sigma_elastic[neutral_index] = self.sigma_return

                return (status, message)


        def read_xsec_electrons_ionization(self, filename, sep, neutral_index, check=False, plot=False):
                """Reads the table of ionization cross section values as a function of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form

                                                        #Comment
                                                        <energy> <separator> <value>    #Comment
                                                        <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                                        

                        sep:            character used to separate values (usually the tab char)

                        neutral_index:  integer value indentifying the type of neutral for which cross section
                                        shuld be read

                        check:          if True, check that all cross section values for energies lower
                                        than ionization threshold are zero

                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation


                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                title    = 'Cross section for e- impact ionization of '+self.names[neutral_index]
                qm_ratio = - ELECTRON_CHARGE / ELECTRON_MASS

                (status, message) = self._read_xsec(filename, sep, qm_ratio, neutral_index, None, plot, title)

                if (status == 0): self.sigma_ionization[neutral_index] = self.sigma_return

                # Check that cross section is zero for energies lower than ionization threshold
                if check:
                        # Calculate speed corresponding to ionization treshold
                        v_min = math.sqrt( 2.0 * self.ionization_energy[neutral_index] * 
                                                                   (-ELECTRON_CHARGE / ELECTRON_MASS) )
                        i=0
                        while(self.sigma_velocities[i]<=v_min): 
                                self.sigma_ionization[neutral_index, i]=0
                                i=i+1

                return (status, message)


        def read_xsec_electrons_dissociation(self, filename, sep, neutral_index, diss_type, check=False, plot=False):
                """Reads the table of dissociation cross section values as a function of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of the file containing cross section values at energies in the form

                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                                

                        sep:            character used to separate values (usually the tab char)

                        neutral_index:  integer value indentifying the type of neutral for which cross section
                                                        shuld be read

                        diss_type:      integer value identifying the type of dissociation process

                        check:          if True, check that all cross section values for energies lower
                                        than dissociation threshold are zero

                        plot:           if True, plot the cross section values as a function of energy,
                                                before and after interpolation

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                title    = 'Cross section for e- impact dissociation of '+ self.names[neutral_index]+' process #'+str(diss_type)
                qm_ratio = - ELECTRON_CHARGE / ELECTRON_MASS

                (status, message) = self._read_xsec(filename, sep, qm_ratio, neutral_index, None, plot, title)

                if (status == 0): self.sigma_dissociation[neutral_index, diss_type] =self.sigma_return

                # Check that cross section is zero for energies lower than dissociation threshold
                if check:
                        # Calculate speed corresponding to dissociation treshold
                        v_min = math.sqrt( 2.0 * self.dissociation_energy[neutral_index, diss_type] * 
                                                                   (-ELECTRON_CHARGE / ELECTRON_MASS) )
                        i=0
                        while(self.sigma_velocities[i]<=v_min): 
                                self.sigma_dissociation[neutral_index, diss_type, i]=0
                                i=i+1                   

                return (status, message)


        def calculate_total_xsec_electrons(self):
                """Calculates main statistical parameters from cross sections.

                        Calculate, for the different processes (elastic scattering, ionization, dissociation types): 
                                - total cross sections
                                - collision frequencies
                                - relative scattering probabilities
                                - probability boundaries
                        (see the module docstring for more information on these quantities and their usage)

                        Parameters
                        ----------

                        Initialized data attributes
                        ---------------------------

                        self.sigma_total_elastic:       sum of elastic scattering cross sections for all neutral types
                        self.sigma_total_ionization:    sum of ionization cross sections for all neutral types
                        self.sigma_total_dissociation:  sum of dissociation cross sections for all neutral types
                                                                and all dissociation processes for each neutral type
                        self.sigma_total_global:        sum of all cross section for all processes

                        self.frequency_elastic:         collision frequencies for elastic scattering
                        self.frequency_ionization:      collision frequencies for ionization scattering
                        self.frequency_dissociation:    collision frequencies for dissociation scattering
                        self.frequency_total_global:    sum of all collision frequencies, for all prossible processes
                        self.frequency_null_collision:  null collision frequency, defined as the difference between
                                                                the maximum value (over all velocities) of the 
                                                                total collision frequency
                                                                and the collision frequency

                        self.p_relative_elastic:        relative probability of elastic scattering
                        self.p_relative_ionization:     relative probability of ionization
                        self.p_relative_dissociation:   relative probabilities dissociation processes
                        self.p_relative_null_collision: relative probability of null collision                                  

                        self.probability_limits:        limits for calculation of scattering type

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                self.sigma_total_elastic        = numpy.zeros( self.n_sigma, 'd')
                self.sigma_total_ionization     = numpy.zeros( self.n_sigma, 'd')
                self.sigma_total_dissociation   = numpy.zeros( self.n_sigma, 'd')
                self.sigma_total_global         = numpy.zeros( self.n_sigma, 'd')

                self.frequency_elastic          = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')
                self.frequency_ionization       = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')
                self.frequency_dissociation     = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES, self.n_sigma), 
                                                               'd', order = 'F')

                self.frequency_total_global     = numpy.zeros( self.n_sigma, 'd')
                self.frequency_null_collision   = numpy.zeros( self.n_sigma, 'd')

                self.p_relative_elastic         = numpy.zeros( (self.types, self.n_sigma), 'd', order='F')
                self.p_relative_ionization      = numpy.zeros( (self.types, self.n_sigma), 'd', order='F')
                self.p_relative_dissociation    = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES, self.n_sigma), 
                                                                'd', order = 'F')
                self.p_relative_null_collision  = numpy.zeros( self.n_sigma, 'd')

                self.probability_limits         = numpy.zeros( (self.types*(2+MAX_DISSOCIATION_TYPES), self.n_sigma), 
                                                                'd', order = 'F')


                # Calculate total cross sections for all processes
                for i in range(self.types):
                        self.sigma_total_elastic    = self.sigma_total_elastic    + self.sigma_elastic[i]
                        self.sigma_total_ionization = self.sigma_total_ionization + self.sigma_ionization[i]
                        for j in range(self.dissociation_types[i]):
                                self.sigma_total_dissociation = self.sigma_total_dissociation + \
                                                                self.sigma_dissociation[i,j]

                # Calculate total global cross-section
                self.sigma_total_global = self.sigma_total_elastic + self.sigma_total_ionization + \
                                          self.sigma_total_dissociation

                # Calculate collision frequencies for all processes
                # f_i(v) = n_i * s_i(v) * v
                for i in range(self.types):
                        self.frequency_elastic[i]    = self.number_density[i] * self.sigma_elastic[i]    * self.sigma_velocities 
                        self.frequency_ionization[i] = self.number_density[i] * self.sigma_ionization[i] * self.sigma_velocities 
                        for j in range(self.dissociation_types[i]):
                                self.frequency_dissociation[i,j] = self.number_density[i] * self.sigma_dissociation[i,j] * \
                                                                   self.sigma_velocities 

                # Calculate total collision frequency 
                # f_tot(v) = sum{f_i(v), i};  f_i(v) = n_i * s_i(v) * v
                for i in range(self.types):
                        self.frequency_total_global = self.frequency_total_global + \
                                                      self.frequency_elastic[i]   + \
                                                      self.frequency_ionization[i]
                        for j in range(self.dissociation_types[i]):
                                self.frequency_total_global = self.frequency_total_global + \
                                                              self.frequency_dissociation[i,j]

                # Calculate the maximum value of the total collision frequency over all velocities
                #   f_max = max{f_tot(v), v}
                self.frequency_total_global_max = self.frequency_total_global.max()

                # Calculate null collision frequency (required to apply the null collision method)
                #   f_null(v) = f_max - f_tot(v)
                self.frequency_null_collision = self.frequency_total_global_max - self.frequency_total_global

                # Calculate relative collision probabilities
                # prel_i(v) = f_i(v) / f_max
                self.p_relative_null_collision = self.frequency_null_collision / self.frequency_total_global_max

                for i in range(self.types):
                        self.p_relative_elastic[i]    = self.frequency_elastic[i]    / self.frequency_total_global_max
                        self.p_relative_ionization[i] = self.frequency_ionization[i] / self.frequency_total_global_max
                        for j in range(self.dissociation_types[i]):
                                self.p_relative_dissociation[i,j]= self.frequency_dissociation[i,j] / self.frequency_total_global_max

                # Calculate probability boundaries
                self.probability_limits[0] = self.p_relative_elastic[0]
                for i in range(1, self.types):
                        self.probability_limits[i] = self.probability_limits[i-1] + \
                                                     self.p_relative_elastic[i]

                for i in range(self.types):
                        self.probability_limits[self.types+i] = self.probability_limits[self.types+i-1] + \
                                                                self.p_relative_ionization[i]
                i = self.types*2
                for j in range(self.types):
                        for k in range(self.dissociation_types[j]):
                                self.probability_limits[i] = self.probability_limits[i-1] + \
                                                             self.p_relative_dissociation[j,k]
                                i = i + 1
                self.n_limits = i
        
                return (status, message)


        def plot_xsec_electrons(self, plot_total=False, plot_frequencies=False, plot_relative=False, plot_boundaries=False,
                                dot_points=False):
                """ Plots cross sections for electrons collisions and related main statistical parameters, 
                    as a function of electron energy.

                        Plots, for the different processes (elastic scattering, ionization, dissociation types): 
                                - total cross sections
                                - collision frequencies
                                - relative scattering probabilities
                                - probability boundaries

                        Parameters
                        ----------

                        plot_total:       if True, plot the total cross section values as a function of energy
                        plot_frequencies: if True, plot the collision frequency values as a function of energy
                        plot_relative:    if True, plot the relative scattering probabilities
                        plot_boundaries:  if True, plot the probability boundaries
                        dot_points:       if True, plot using dots instead of markers

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking

                if (not gnuplot_installed): return

                if dot_points: plot_style = 'dots'
                else:          plot_style = 'points'
               
                # Plot total cross-sections
                if plot_total:
                        self.plot_list = []
                        # Prepare graph
                        self.cross_section_graph = new_plot(title='Cross sections for e- impact',
                                                            xlabel='Energy / eV',
                                                            ylabel='Cross section / m**2',
                                                            logx=True,
                                                            logy=True,
                                                            format_x='%g',
                                                            format_y='%g',
                                                            grid=True,
                                                            style=plot_style,
                                                            persistence=XSEC_PLOT_PERSIST,
                                                            redirect_output=True)
                        # Plot total global cross section
                        self.plot_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_global, 
                                                 'total global'] )
                        # Plot total cross sections for each process type
                        self.plot_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_elastic,
                                                 'total elastic' ] )
                        self.plot_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_ionization,
                                                 'total ionization'] )
                        self.plot_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_dissociation,
                                                 'total dissociation' ] )

                        # Plot cross sections for each process type and gas type
                        for i in range(self.types):
                                self.plot_list.append( [ self.sigma_energies(),
                                                         self.sigma_elastic[i],
                                                         'elastic ' + self.names[i] ] )
                        for i in range(self.types):
                                self.plot_list.append( [ self.sigma_energies(),
                                                         self.sigma_ionization[i],
                                                         'ionization ' + self.names[i] ] )
                        for i in range(self.types):
                                for j in range(self.dissociation_types[i]):
                                        self.plot_list.append( [ self.sigma_energies(),
                                                                 self.sigma_dissociation[i,j],
                                                                 'dissociation ' + self.names[i] + ' # ' + str(j) ] )

                        plot_curves(self.cross_section_graph, self.plot_list)
                        
                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.cross_section_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.cross_section_graph)
                
                # Plot collision frequencies                
                if plot_frequencies:
                        self.plot_list = []
                        self.coll_freq_graph = new_plot(title='Collision frequencies',
                                                        xlabel='Energy / eV',
                                                        ylabel='Frequency / Hz',
                                                        format_x='%g',
                                                        format_y='%g',
                                                        grid=True,
                                                        style=plot_style,
                                                        persistence=XSEC_PLOT_PERSIST,
                                                        redirect_output=True)                        
                        for i in range(self.types):
                                self.plot_list.append( [ self.sigma_energies(),
                                                         self.frequency_elastic[i],
                                                         'elastic ' + self.names[i] ] )
                                self.plot_list.append( [ self.sigma_energies(),
                                                         self.frequency_ionization[i],
                                                         'ionization ' + self.names[i] ] )
                                for j in range(self.dissociation_types[i]):
                                        self.plot_list.append( [ self.sigma_energies(),
                                                                 self.frequency_dissociation[i,j],
                                                                 'dissociation ' + self.names[i] + ' # ' + str(j) ] )
                        self.plot_list.append( [ self.sigma_energies(),
                                                 self.frequency_total_global,
                                                 'total' ] )
                        self.plot_list.append( [ self.sigma_energies(),
                                                 self.frequency_null_collision,
                                                 'null collision' ] )
                        self.plot_list.append( [ self.sigma_energies(),
                                                 self.frequency_total_global_max*numpy.ones(self.n_sigma),
                                                 'max' ] )

                        plot_curves(self.coll_freq_graph, self.plot_list)
                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.coll_freq_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.coll_freq_graph)
                        
                # Plot relative collision frequencies
                if plot_relative:
                        self.plot_list = []
                        self.rel_coll_freq_graph = new_plot(title='Relative collision frequencies',
                                                            xlabel='Energy / eV',
                                                            ylabel='Relative probability / %',
                                                            logx=True,
                                                            grid=True,
                                                            style=plot_style,
                                                            persistence=XSEC_PLOT_PERSIST,
                                                            redirect_output=True)                        
                        for i in range(self.types):
                                self.plot_list.append( [ self.sigma_energies(),
                                                         self.p_relative_elastic[i]*100,
                                                         'elastic '+self.names[i] ] )
                                self.plot_list.append( [ self.sigma_energies(),
                                                         self.p_relative_ionization[i]*100,
                                                         'ionization '+self.names[i] ] )
                                for j in range(self.dissociation_types[i]):
                                        self.plot_list.append( [ self.sigma_energies(),
                                                                 self.p_relative_dissociation[i,j]*100,
                                                                 'dissociation ' + self.names[i] + ' # ' + str(j) ] )
                        self.plot_list.append( [ self.sigma_energies(),
                                                 self.p_relative_null_collision*100,
                                                 'null collision' ] )

                        plot_curves(self.rel_coll_freq_graph, self.plot_list)
                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.rel_coll_freq_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.rel_coll_freq_graph)
                        
                # Plot probability boundaries
                if plot_boundaries:
                        self.plot_list = []
                        self.boundaries_graph = new_plot(title='Probability boundaries',
                                                         xlabel='Energy / eV',
                                                         ylabel='Probability limit',
                                                         logx=True,
                                                         grid=True,
                                                         style=plot_style,
                                                         persistence=XSEC_PLOT_PERSIST,
                                                         redirect_output=True)
                        for i in range(self.n_limits):
                                self.plot_list.append( [ self.sigma_energies(),
                                                         self.probability_limits[i],
                                                         'boundary #'+str(i) ] )

                        plot_curves(self.boundaries_graph, self.plot_list)
                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.boundaries_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.boundaries_graph)
                        
                return status, message


        # +-------------------------------------------------------------+
        # | Methods to read cross sections for ions/neutrals scattering |
        # +-------------------------------------------------------------+

        def read_xsec_ions_elastic(self, filename, sep, ion_index, target_index, plot=False):
                """Reads the table of cross sections values for ions-neutrals elastic scattering as a function 
                   of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form

                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                        
                        sep:            character used to separate values (usually the tab char)
                        ion_index:      integer value indentifying the type of ion (i.e. the type of neutral that was ionized) 
                                        that elastically scatters with the neutral
                        target_index:   integer value indentifying the type of neutral target
                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                title    = 'Cross section for elastic scattering of ' + self.names[ion_index] + '+ on ' + self.names[target_index]
                qm_ratio = -ELECTRON_CHARGE / (self.mass[ion_index] * ATOMIC_UNIT_MASS)

                (status, message) = self._read_xsec(filename, sep, qm_ratio, target_index, ion_index, plot, title)

                if (status == 0): self.sigma_elastic_ions[ion_index, target_index] = self.sigma_return

                return (status, message)


        def read_xsec_ions_charge_exchange(self, filename, sep, ion_index, target_index, plot=False):
                """Reads the table of cross sections values for ions-neutrals charge exchange scattering 
                   as a function of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form

                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                        
                        sep:            character used to separate values (usually the tab char)
                        ion_index:      integer value indentifying the type of ion (i.e. the type of neutral that was ionized) 
                                        that scatters with the neutral
                        target_index:   integer value indentifying the type of neutral target


                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                title    = 'Cross section for charge exchange scattering of ' + self.names[ion_index] + '+ on ' + \
                           self.names[target_index] 
                qm_ratio = -ELECTRON_CHARGE / (self.mass[ion_index] * ATOMIC_UNIT_MASS)

                (status, message) = self._read_xsec(filename, sep, qm_ratio, target_index, ion_index, plot, title)

                if (status == 0): self.sigma_charge_exchange_ions[ion_index,target_index] = self.sigma_return

                return (status, message)


        def calculate_total_xsec_ions(self):
                """Calculates main statistical parameters from cross sections for ion impact processes.

                        Calculate, for the different processes (elastic scattering, ionization, dissociation types): 
                                - total cross sections
                                - collision frequencies
                                - relative scattering probabilities
                                - probability boundaries
                        (see the module docstring for more information on these quantities and their usage)

                        Parameters
                        ----------

                        Initialized data attributes
                        ---------------------------

                        self.sigma_total_global_ions:        sum of all cross section for all processes

                        self.frequency_elastic_ions:         collision frequencies for elastic scattering
                        self.frequency_charge_exchange_ions: collision frequencies for charge exchange scattering
                        self.frequency_total_global_ions:    sum of all collision frequencies, for all prossible processes
                        self.frequency_null_collision_ions:  null collision frequency, defined as the difference between
                                                                the maximum value (over all velocities) of the 
                                                                total collision frequency
                                                                and the collision frequency

                        self.p_relative_elastic_ions:        relative probability of elastic scattering
                        self.p_relative_exchange_ions:       relative probability of ionization
                        self.p_relative_null_collision:      relative probability of null collision                                     

                        self.probability_limits_ions:        limits for calculation of scattering type

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                self.sigma_total_global_ions            = numpy.zeros( (self.types, self.n_sigma_ions), 'd', order='F')

                self.frequency_elastic_ions             = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')
                self.frequency_charge_exchange_ions     = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')
                self.frequency_total_global_ions        = numpy.zeros( (self.types, self.n_sigma_ions), 'd', order='F')
                self.frequency_null_collision_ions      = numpy.zeros( (self.types, self.n_sigma_ions), 'd', order='F')

                self.frequency_total_global_max_ions    = numpy.zeros( self.types, 'd')

                self.p_relative_elastic_ions            = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')
                self.p_relative_charge_exchange_ions    = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')
                self.p_relative_null_collision_ions     = numpy.zeros( (self.types, self.n_sigma_ions), 'd', order='F')

                self.probability_limits_ions            = numpy.zeros( (self.types, 2*self.types, self.n_sigma_ions), 
                                                                      'd', order = 'F')

                # Calculate total global cross-section
                for ion in range(self.types):
                        for target in range(self.types):
                                self.sigma_total_global_ions[ion] = self.sigma_total_global_ions[ion] + \
                                                                    self.sigma_elastic_ions[ion,target] + \
                                                                    self.sigma_charge_exchange_ions[ion,target]

                # Calculate collision frequencies for all processes
                # f_i(v) = n_i * s_i(v) * v
                for ion in range(self.types):
                        for target in range(self.types):
                                self.frequency_elastic_ions[ion,target]         = self.number_density[target] * \
                                                                                  self.sigma_elastic_ions[ion,target] * \
                                                                                  self.sigma_velocities_ions[ion]
                                self.frequency_charge_exchange_ions[ion,target] = self.number_density[target] * \
                                                                                  self.sigma_charge_exchange_ions[ion,target] * \
                                                                                  self.sigma_velocities_ions[ion] 

                # Calculate total collision frequency 
                # f_tot(v) = sum{f_i(v), i}
                for ion in range(self.types):
                        for target in range(self.types):
                                self.frequency_total_global_ions[ion] = self.frequency_total_global_ions[ion] + \
                                                                        self.frequency_elastic_ions[ion,target] + \
                                                                        self.frequency_charge_exchange_ions[ion,target]

                # Calculate the maximum value of the total collision frequency over all velocities
                #   f_max = max{f_tot(v), v}
                for ion in range(self.types):
                        self.frequency_total_global_max_ions[ion] = self.frequency_total_global_ions[ion].max()

                # Calculate null collision frequency (required to apply the null collision method)
                #   f_null(v) = f_max - f_tot(v)
                for ion in range(self.types):
                        self.frequency_null_collision_ions[ion] = self.frequency_total_global_max_ions[ion] - \
                                                                  self.frequency_total_global_ions[ion]

                # Calculate relative collision probabilities
                # prel_i(v) = f_i(v) / f_max
                for ion in range(self.types):
                        self.p_relative_null_collision_ions[ion] = self.frequency_null_collision_ions[ion] / \
                                                                   self.frequency_total_global_max_ions[ion]
                        for target in range(self.types):
                                self.p_relative_elastic_ions[ion,target]         = self.frequency_elastic_ions[ion,target] / \
                                                                                   self.frequency_total_global_max_ions[ion]
                                self.p_relative_charge_exchange_ions[ion,target] = self.frequency_charge_exchange_ions[ion,target] /\
                                                                                   self.frequency_total_global_max_ions[ion]

                # Calculate probability boundaries #######################################################
                for ion in range(self.types):
                        self.probability_limits_ions[ion][0] = self.p_relative_elastic_ions[ion][0]
                        for i in range(1, self.types):
                                self.probability_limits_ions[ion][i] = self.probability_limits_ions[ion][i-1] + \
                                                                       self.p_relative_elastic_ions[ion][i]
                        for i in range(self.types):
                                self.probability_limits_ions[ion][self.types+i] = self.probability_limits_ions[ion][self.types+i-1]+\
                                                                                  self.p_relative_charge_exchange_ions[ion][i]
                self.n_limits_ions = 2*self.types
        
                return (status, message)


        def plot_xsec_ions(self, ion_type, plot_total=False, plot_frequencies=False, plot_relative=False, plot_boundaries=False,
                           dot_points=False):
                """Plots cross sections for ion collisions and related main statistical parameters, as a function of electron energy.

                        Plots, for the different processes (elastic scattering, ionization, dissociation types): 
                                - total cross sections
                                - collision frequencies
                                - relative scattering probabilities
                                - probability boundaries

                        Parameters
                        ----------

                        ion_type:         type of ion for which the quantities will be plotted
                        plot_total:       if True, plot the total cross section values as a function of energy
                        plot_frequencies: if True, plot the collision frequency values as a function of energy
                        plot_relative:    if True, plot the relative scattering probabilities
                        plot_boundaries:  if True, plot the probability boundaries

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking

                if (not gnuplot_installed): return

                if dot_points: plot_style = 'dots'
                else:          plot_style = 'points'               

                # Plot total cross-sections
                if plot_total:
                        self.plot_list = []
                        # Prepare graph
                        self.cross_section_graph = new_plot(title='Cross sections for ' + self.names[ion_type] + '+ impact',
                                                            xlabel='Energy / eV',
                                                            ylabel='Cross section / m**2',
                                                            logx=True,
                                                            logy=True,
                                                            grid=True,
                                                            style=plot_style,
                                                            persistence=XSEC_PLOT_PERSIST,
                                                            redirect_output=True)
                        # Plot total global cross section
                        self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                 self.sigma_total_global_ions[ion_type],
                                                 'total global' ] )
                        # Plot cross sections for each process type
                        for target in range(self.types):
                                self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                         self.sigma_elastic_ions[ion_type,target],
                                                         'elastic ' + self.names[ion_type] + '+ on ' + self.names[target] ] )
                                self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                         self.sigma_charge_exchange_ions[ion_type,target],
                                                         'charge exchange '+self.names[ion_type]+'+ on '+ self.names[target] ] )
                                
                        plot_curves(self.cross_section_graph, self.plot_list)
                        
                        plot_close(self.cross_section_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.cross_section_graph)
                                
                # Plot collision frequencies
                if plot_frequencies:
                        self.plot_list = []
                        self.coll_freq_graph = new_plot(title='Collision frequencies for ' + self.names[ion_type] + '+ impact',
                                                        xlabel='Energy / eV',
                                                        ylabel='Frequency / Hz',
                                                        logx=True,
                                                        logy=True,
                                                        format_x='%g',
                                                        format_y='%g',
                                                        grid=True,
                                                        style=plot_style,
                                                        persistence=XSEC_PLOT_PERSIST,
                                                        redirect_output=True)
                        
                        for target in range(self.types):
                                self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                         self.frequency_elastic_ions[ion_type,target],
                                                         'elastic ' + self.names[ion_type] + '+ on ' + self.names[target] ] )
                                self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                         self.frequency_charge_exchange_ions[ion_type,target],
                                                         'charge exchange ' + self.names[ion_type] + '+ on ' + self.names[target] ] )
                        self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                 self.frequency_null_collision_ions[ion_type],
                                                 'null collision ' + self.names[ion_type] + '+' ] )
                        self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                 self.frequency_total_global_ions[ion_type],
                                                 'total ' + self.names[ion_type] + '+' ] )
                        self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                 self.frequency_total_global_max_ions[ion_type]*numpy.ones(self.n_sigma_ions),
                                                 'max ' + self.names[ion_type] + '+' ] )

                        plot_curves(self.coll_freq_graph, self.plot_list)

                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.coll_freq_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.coll_freq_graph)

                # Plot relative collision frequencies
                if plot_relative:
                        self.plot_list = []
                        self.cross_sect_rel_graph = new_plot(title='Relative collision frequencies for ' + self.names[ion_type] + '+ impact',
                                                             xlabel='Energy / eV',
                                                             ylabel='Relative probability / %',
                                                             logx=True,
                                                             grid=True,
                                                             style=plot_style,
                                                             persistence=XSEC_PLOT_PERSIST,
                                                             redirect_output=True)
                        for target in range(self.types):
                                self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                         self.p_relative_elastic_ions[ion_type,target]*100,
                                                         'elastic '+self.names[ion_type]+'+ on '+self.names[target] ] )
                                self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                         self.p_relative_charge_exchange_ions[ion_type,target]*100,
                                                         'charge exchange ' + self.names[ion_type] + '+ on ' + self.names[target] ] )
                        self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                 self.p_relative_null_collision_ions[ion_type]*100,
                                                 'null collision ' + self.names[ion_type] + '+' ] )

                        plot_curves(self.cross_sect_rel_graph, self.plot_list)
                        
                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.cross_sect_rel_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.cross_sect_rel_graph)

                # Plot probability boundaries
                if plot_boundaries:
                        self.plot_list = []
                        self.boundaries_graph = new_plot(title='Probability boundaries for ' + self.names[ion_type] + '+ impact',
                                                         xlabel='Energy / eV',
                                                         ylabel='Probability limit',
                                                         logx=True,
                                                         grid=True,
                                                         style=plot_style,
                                                         persistence=XSEC_PLOT_PERSIST,
                                                         redirect_output=True)
                        for i in range(self.n_limits_ions):
                                self.plot_list.append( [ self.sigma_energies_ions(ion_type),
                                                         self.probability_limits_ions[ion_type][i],
                                                         'boundary #'+str(i) ] )
                                
                        plot_curves(self.boundaries_graph, self.plot_list)
                        
                        # Shut down gnuplot process, window remain visible due to persist option        
                        plot_close(self.boundaries_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.boundaries_graph)

                return status, message


        # +----------------------------------------------------------------------+
        # | Methods to read cross sections for electrons recombination with ions |
        # +----------------------------------------------------------------------+

        def read_xsec_ele_ion_recomb(self, filename, sep, ion_index, rec_type=None, diss_type=None, plot=False):
                """Reads the table of cross section values for recombination of electrons with ions.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form

                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                        
                        sep:            character used to separate values (usually the tab character)
                        ion_index:      integer value indentifying the type of ion (must be a molecule) 
                                        with which the elctron recombine
                        rec_type:       type of recombination process
                                        'dissociative' -> dissociative recombination of e- with a molecular ion
                                        '3body'        -> recombination of e- with ion, with transfer of the released energy 
                                                          to another e-
                                        'radiative'    -> recombination of e- with ion, with radiative release of excess energy

                        diss_type:      integer value indentifying the type of dissociative recombination process to which
                                        the cross section is referred (only for dissociative recombination processes)
                                        (the number of dissociation types for each molecule type is defined 
                                        in the "neutrals.csv" file)
                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation


                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                         3 = ion index is outside allowed bounds
                                         4 = unspecified recombination type
                                         5 = unknown recombination type
                                         6 = dissociation process index is missing
                                         7 = dissociation process index is out of bounds
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                if ( (ion_index < 0) or (ion_index > self.types-1)):
                        status = 3
                        message = 'ion type ' + str(ion_index) + ' is out of bounds'
                        return(status, message)
                if (rec_type is None):
                        status = 4
                        message = 'unspecified recombination type'
                        return(status, message)
                if (rec_type not in ('dissociative', '3body', 'radiative')):
                        status = 5
                        message = 'unknown recombination type \"' + str(rec_type) + '\"'
                        return(status, message)
                if (rec_type == 'dissociative'):
                        if (diss_type is None):
                                status = 6
                                message = 'dissociation process index is missing'
                                return (status, message)
                        if (diss_type > self.dissociation_types[ion_index]):
                                status = 7
                                message = 'dissociation process index ' + str(diss_type) + ' is out of bounds'
                                return (status, message)

                title    = 'Cross section for ' + rec_type + ' recombination of e- with ' + self.names[ion_index]
                if (rec_type =='dissociative'): 
                        title = title + ' (process #' + str(diss_type) + ')'
                qm_ratio = - ELECTRON_CHARGE / ELECTRON_MASS

                (status, message) = self._read_xsec(filename, sep, qm_ratio, ion_index, None, plot, title)

                if (status == 0): 
                        if (rec_type == 'dissociative'): self.sigma_recombination_diss[ion_index,diss_type] = self.sigma_return
                        if (rec_type == '3body'):        self.sigma_recombination_3body[ion_index]          = self.sigma_return
                        if (rec_type == 'radiative'):    self.sigma_recombination_rad[ion_index]            = self.sigma_return

                return (status, message)


        def calculate_total_xsec_ele_ion_recomb(self):
                """Calculates main statistical parameters from cross sections.

                        Calculate, for the different processes of electron-ion recombination:
                                - total cross sections
                                - recombination rate coefficients (non averaged), i.e. product cross section by electron velocity
                        (see the module docstring for more information on these quantities and their usage)

                        When calculating the recombination process for an electron with a given ion type, 
                        the background density involved in the calculation is the density of that ion type.
                        Since it is not constant [*], it is not possible to pre-calculate 
                        recombination frequencies, nor recombination rates. They must be calculated during simulation
                        at each timestep based on the ion density at that time.

                        [*] Instead, neutral density can be considered appoximately constant, if the ionization degree is very low
                        so collision frequencies with neutrals can be pre-calculated.

                        Parameters
                        ----------

                        Initialized data attributes
                        ---------------------------

                        self.ratecoeff_recomb_diss:     rate coefficients (sigma*v) for dissociative recombination processes 
                                                             of each ion type

                        self.sigma_total_recomb_diss:   sum of dissociative recombination cross sections for all neutral types
                                                             and all dissociation processes for each neutral type

                        self.sigma_total_global_recomb: sum of all cross section for all recombination processes

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                self.sigma_total_recombination_diss    = numpy.zeros( self.n_sigma, 'd' )
#               self.sigma_total_recombination_3body   = numpy.zeros( self.n_sigma, 'd' )
#               self.sigma_total_recombination_rad     = numpy.zeros( self.n_sigma, 'd' )
                self.sigma_total_global_recombination  = numpy.zeros( self.n_sigma, 'd' )
                self.ratecoeff_recomb_diss             = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES, self.n_sigma), 
                                                                      'd', order = 'F')
                # Calculate total cross sections for each recombination process
                # (at the moment only dissociative recombination is considered)
                for ion_index in range(self.types):
                        for diss_type in range(self.dissociation_types[ion_index]):
                                self.sigma_total_recombination_diss = self.sigma_total_recombination_diss + \
                                                                      self.sigma_recombination_diss[ion_index,diss_type]

                # Calculate total global cross-section, for all the possible recombination types
                # (at the moment only dissociative recombination is considered,
                # so it's equal to total cross section for that type of recombination)
                self.sigma_total_global_recombination  = self.sigma_total_recombination_diss
#                                                       + self.sigma_total_recombination_3body
#                                                       + self.sigma_total_recombination_rad

                # Calculate rate coefficients (not averaged) for each recombination process
                # (at the moment only dissociative recombination is considered)
                # rate coefficient is the product cross section by velocity -> sigma(v) * v
                for ion_index in range(self.types):
                        for diss_type in range(self.dissociation_types[ion_index]):
                                self.ratecoeff_recomb_diss[ion_index,diss_type]= self.sigma_recombination_diss[ion_index,diss_type]*\
                                                                                 self.sigma_velocities
#                        self.ratecoeff_recomb_3body[ion_index] = self.sigma_recombination_3body[ion_index] * self.sigma_velocities     
#                        self.ratecoeff_recomb_rad[ion_index]   = self.sigma_recombination_rad[ion_index]   * self.sigma_velocities

                return (status, message)


        def plot_xsec_ele_ion_recomb(self, plot_total=False, plot_rates=False, dot_points=False):
                """Plots cross sections for electrons collisions and related main statistical parameters, 
                   as a function of electron energy.

                        Plots, for the different processes (elastic scattering, ionization, dissociation types): 
                                - total cross sections
                                - collision rates 

                        Parameters
                        ----------

                        plot_total:       if True, plot the total cross section values as a function of energy
                        plot_rates:       if True, plot the collision rate values as a function of energy

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking

                if (not gnuplot_installed): return

                if dot_points: plot_style = 'dots'
                else:          plot_style = 'points'                

                # Plot total cross-sections
                if plot_total:
                        self.plot_list = []
                        self.el_ion_recomb_graph = new_plot(title='Cross sections for e-/ion recombination',
                                                            xlabel='Energy / eV',
                                                            ylabel='Cross section / m**2',
                                                            logx=True,
                                                            logy=True,
                                                            format_x='%g',
                                                            format_y='%g',
                                                            grid=True,
                                                            style=plot_style,
                                                            persistence=XSEC_PLOT_PERSIST,
                                                            redirect_output=True)                                                      
                        self.plot_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_global_recombination, 
                                                 'total global' ] )

                        # Plot total cross sections for each process type
                        self.plot_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_recombination_diss, 
                                                 'total dissociative recombination' ] )
                        # Plot cross sections for each process type and gas type
                        for i in range(self.types):
                                for j in range(self.dissociation_types[i]):
                                        self.plot_list.append( [ self.sigma_energies(),
                                                                 self.sigma_recombination_diss[i,j],
                                                                 'dissociative recombination ' + self.names[i] + ' # '+str(j) ] )
                                        
                        plot_curves(self.el_ion_recomb_graph, self.plot_list)
                                         
                        plot_close(self.el_ion_recomb_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.el_ion_recomb_graph)

                # Plot collision rates
                if plot_rates:
                        self.plot_list = []
                        self.coll_rates_graph = new_plot(title='Collision rate coefficients for e-/ion recombination',
                                                         xlabel='Energy / eV',
                                                         ylabel='Collision rate coefficients / m**3 s**-1',
                                                         logx=True,
                                                         logy=True,
                                                         format_x='%g',
                                                         format_y='%g',
                                                         grid=True,
                                                         style=plot_style,
                                                         persistence=XSEC_PLOT_PERSIST,
                                                         redirect_output=True)                                                       
                        for i in range(self.types):
                                for j in range(self.dissociation_types[i]):
                                        self.plot_lsit.append( [ self.sigma_energies(),
                                                                 self.ratecoeff_recomb_diss[i,j],
                                                                 ( 'dissociative recombination '
                                                                   + self.names[i]
                                                                   + ' # '
                                                                   + str(j) ) ] )

                        plot_curves(self.coll_rates_graph, self.plot_list)

                        plot_close(self.coll_rates_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.coll_rates_grap)

                return status, message


        def initialize_savefile(self, filename_neutral, append=False, sep='\t', ext='.csv'):
                """Initializes the file to which simulation data about neutral particles will be saved

                   Parameters
                   ----------
                   filename_neutral:      name of the file to which the simulation quantities will be saved
                   sep:                   character or sequence to separate data columns
                   ext:                   extension to give to filename (deafult is '.csv')

                   Initialized data attributes
                   ---------------------------

                   append:                if True, open the file for appendig new data to the end
                                          header will not be re-written
                   self.sep:              character or sequence to separate data columns
                   self.f_neutral:        file to which values of several quantities will be saved
                """

                self.sep = sep
                self.filename = str(filename_neutral) + ext
                
                # Open data file where to save time evolution of mean values
                if append:
                        data_file = open(self.filename,'a')
                else:
                        data_file = open(self.filename,'w')
                        # Write header
                        data_file.write('\"t[ns]\"' + self.sep )                
                        for i in range(self.types):
                                for j in range(self.dissociation_types[i]):
                                        data_file.write('\"Diss_rate ' + self.names[i] + '#'+str(j) + \
                                                          ' [m**-3*s**-1]\"' + self.sep   )
                                        data_file.write('\"Diss_rate const ' + self.names[i] + '#'+str(j) + \
                                                          ' [m**3*s**-1]\"' + self.sep   )
                        data_file.write(EOL)
                data_file.close()


        def save_data_to_files(self, time):
                """Saves actual data values to files """

                data_file = open(self.filename,'a')
                data_file.write( str(1E9*time) + self.sep )                
                for i in range(self.types):
                        for j in range(self.dissociation_types[i]):
                                data_file.write( str(self.dissociation_rate[i][j]) + self.sep )
                                data_file.write( str(self.dissociation_rate_constant[i][j]) + self.sep )
                data_file.write(EOL)
                data_file.close()



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Parameters used in several modules of this package.

    This module contains some parameters, which are used in several modules
    and subpackages.
"""

# Parameters used for stadard behavior when plotting via matplotlib or gnuplot
#PLOT_INTERFACE  = 'gnuplot'
PLOT_INTERFACE  = 'pylab'
PLOT_SYMBOL     = '+'
PLOT_COLOR      = 'red'
PLOT_SYMBOLS    = [ '.',     'o',    '+',    'x',     '^',    'v',       '>',   '<',      'D'      ]
N_PLOT_SYMBOLS  = len(PLOT_SYMBOLS)
PLOT_COLORS     = [ 'black', 'grey', 'blue', 'green', 'cyan', 'magenta', 'red', 'orange', 'yellow' ]
N_PLOT_COLORS   = len(PLOT_COLORS)
FILL_COLOR      = 'grey'
PLOT_LINE       = 'None'

# Standard answer for error checking
OK	        = 'Ok'

# Standard characters used to separate data rows and columns
EOL		= '\n'	   # End-of-line character
SEP		= '\t'	   # Separator character for data input/output files
ZERO		= 1.0E-9

GPL_MESSAGE     = ( '\n'
                    + 'COPYRIGHT 2008 Pietro Mandracci\n'
                    + '\n'
                    + 'This program is free software: you can redistribute it and/or modify\n'
                    + 'it under the terms of the GNU General Public License as published by\n'
                    + 'the Free Software Foundation, either version 3 of the License, or\n'
                    + '(at your option) any later version.\n'
                    + '\n'
                    + 'This program is distributed in the hope that it will be useful,\n'
                    + 'but WITHOUT ANY WARRANTY; without even the implied warranty of\n'
                    + 'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n'
                    + 'GNU General Public License for more details.\n'
                    + '\n'
                    + 'You should have received a copy of the GNU General Public License\n'
                    + 'along with this program.  If not, see <http://www.gnu.org/licenses/>.\n'
                    + '\n' )



# COPYRIGHT 2020 Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Simulation of a capacitively coupled plasma discharge: numerical kernel """

from pysica.plasmapro.ccpla_defaults import *
from pysica.plasmapro.discharge import particle_mover


def kernel(signal_connection, time_connection, data_connection, debug=False):

    """ This function runs parallel to the GUI and calls the particle mover """

    if debug: print('[Kernel]-> ***INSTALLED***\n')

    stay_alive = True
    while stay_alive:
        # Wait until a message arrives on the message pipe
        if debug: print('[Kernel]-> waiting for message [start/quit]')
        message = signal_connection.recv()
        if debug: print('[Kernel]-> message is \"' + message + '\"')        
        # If message is 'quit', exit from main loop
        if (message == 'quit'):
            stay_alive = False            
        # If message is 'start', begin the simulation (if it is anything else, do nothing)
        elif (message == 'start'):            
            if debug: print('[Kernel]-> waiting for initialized data\n')
            (charges, neutrals, ccp, parameters, options) = data_connection.recv()
            if debug: print('[Kernel]-> initialized data read\n')            
            # Start simulation loop
            run_simulation = True
            while run_simulation:                
                # Read the iteration time requested
                if debug: print('[Kernel]-> waiting for iteration time')
                time = time_connection.recv()
                parameters.dt_output = time                  
                if debug: print('[Kernel]-> iteration time is ' +str(parameters.dt_output))
                
                # Call the particle mover
                if debug: print('[Kernel]-> calling particle mover')
                particle_mover.move_particles(charges, neutrals, ccp, parameters, options)
                if (charges.n_active(0) <= 0):
                    if debug: print('[Kernel]-> no more electrons') 
                    run_simulation = False
                elif ( (parameters.sim_duration > 0) and (charges.time >= parameters.sim_duration) ):
                    if debug: print('[Kernel]-> time completed') 
                    run_simulation = False
                # Send to pipes signal of completed iteration and data
                if debug: print('[Kernel]-> sending message [completed]')
                signal_connection.send('completed')
                if debug: print('[Kernel]-> sending data')                        
                data_connection.send((charges, neutrals, ccp, parameters, options))
                if run_simulation:
                    # Read a signal 
                    if debug: print('[Kernel]-> waiting for message [continue/stop]')
                    message = signal_connection.recv()
                    if debug: print('[Kernel]-> message is \"' + message + '\"')
                    if (message == 'stop'):
                        if debug: print('[Kernel]-> simulation stopped')
                        run_simulation = False
                if debug: print('[Kernel]-> end loop (run_simulation = ' + str(run_simulation) +')\n')
            
    if debug: print('[Kernel]-> ***QUITTING***\n')



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" PYthon tools for SImulation and CAlculus: calculation of optical absorption in thin films

    This module contains some functions that may be useful for the calculation
    of some optical quantities (such as reflactive index or absorption coefficient) in thin films.

    Documentation is also available in the docstrings.
"""

# +-------------------------+
# | Import required modules |
# +-------------------------+

# Mudules from the standard Python library
import math

# Modules from the Python community
import numpy

#
# All functions, except "alpha_brutal" and "n_brutal", from: R. Swanepoel, J. Phys. E: Sci. Instrum, vol. 16 (1983) pp. 1214-1222
#
# s    -> refractive index of the substrate
# n    -> refractive index of the film
# t    -> film thickness
# TMax -> value of transmittance maxima envelope
# Tmin -> value of transmittance minima envelope
# Ta   -> value of geometric average of transmittance maxima and minima envelope 


# +------------------------------+
# | Refractive index calculation |
# +------------------------------+

def n_transparent_1(s, Tmin):
        M  = 2*s / Tmin  -  (s*s  +  1)  /  2
        return numpy.sqrt( M + numpy.sqrt(M*M - s*s) )

# Valid in transparent region using Ta
def n_transparent_2(s, Ta):
        H  = 4*s*s / ( Ta*Ta * (s*s + 1) ) - (s*s + 1) / 2
        return numpy.sqrt( H + numpy.sqrt(H*H - s*s) )

def n_lowabsorption(s, TMax, Tmin):
        N  = 2*s * (TMax-Tmin) / (TMax*Tmin) + (s*s + 1) / 2
        return numpy.sqrt( N + numpy.sqrt(N*N - s*s) )

# Calculate n from the distance between adjacent maxima/minima of transmission or reflection spectrum
# t  -> film thickness in nm
# dE -> energy separation between Maxima (or minima, since they should be equal) in eV
def n_brutal(dE, t):
        return 1239.84 / (2*dE*t) 

# +------------------------------------+
# | Absorption coefficient calculation |
# +------------------------------------+

# Calculate the absorption coefficient
# t  -> film thickness in nm
# n  -> refrative index of the thin film
# s  -> refrative index of the substrate
# Ta -> optical transmittance (fringes free, i.e. geometric average of maxima and minima envelopes) 
# T0 -> 1-R, where R is the optical reflectance


def alpha_A3(s, n, Ta, t):
        R1 = ( (1-n) / (1+n) )**2
        R2 = ( (n-s) / (n+s) )**2
        R3 = ( (s-1) / (s+1) )**2
        P  = (R1-1) * (R2-1) * (R3-1)   
        
        Q  = 2 * Ta * (R1*R2 + R1*R3 - 2*R1*R2*R3)
        x  = ( P + numpy.sqrt( P*P + 2*Q*Ta*(1-R2*R3) ) ) / Q
        return -numpy.log(x) / t


def alpha_19(s, n, Ta, t):
        G1 = 128*n**4*s*s / Ta**2
        G2 = n*n * (n*n - 1)**2 * (s*s - 1)**2
        G3 = (n*n - 1)**2 * (n*n - s*s)**2
        G  = G1 + G2 + G3       
        
        x0 = numpy.sqrt( G*G - (n*n - 1)**6 * (n*n-s**4)**2 )
        x  = numpy.sqrt(G - x0) / ((n - 1)**3 * (n - s*s))
        return -numpy.log(x) / t


def alpha_brutal(T0, Ta, t):
        return numpy.log(T0/Ta) / t



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Tools for the simulation of plasma discharges: fortran module

    This sub-package contains some modules, compiled from Fortran using f2py, which can be used for simulation 
    of the motion of particles (mainly electrons) in cold plasmas using leap-frog and PIC schemes

    * modules

            - fmodule	-> functions used to simulate the motion of charged particles in a cold plasma discharge

    Documentation is also available in the docstrings.
"""



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" PYthon tools for SImulation and CAlculus: functions for the expression of quantities in SI units using prefixes.

    This module contains some functions which allow to manage the use prefixes when printing a quantity 
    in SI units, just like in the following situation:
    * you want to print a quantity expressed in a certain SI unit (e.g. an energy expressed in joules)
    * the quantity may vary among several orders of magnitude (e.g. from 1E-15 J to 1E6 J)
    * you want to express the quantity using SI prefixies (e.g. 1 fJ and 1 MJ instead of 1E-15 J and 1E6 J)
    * you may also want to limit the number of significant digits to be shown (e.g. 1.15 nJ instead of 
      1.149456E-9 J)

    The most straightforward use of this module is to use the function print_unit, 
    giving to it a numerical value and a string that contains the symbol of the SI unit 
    in which the value is expressed.  
    The function will return a string with the value expressed in the unit, using prefixies, and limiting the 
    number of significant digits if requested:
        print_unit(1.123E-3, 'J')    -> '1.123 mJ'
        print_unit(1123456, 'J')     -> '1.123456 MJ'
        print_unit(1123456, 'J', 2)  -> '1.1 MJ'

    As an option, you may also request a check that the string you provide is a SI unit.

    Documentation is also available in the docstrings.
"""

#+-------------------------+
#| Import required modules |
#+-------------------------+

import math


#+---------------------+
#| Numerical constants |
#+---------------------+

units = (
    # Fundamental SI units
    'm',
    'kg',
    's',
    'A',
    'K',
    'mol',
    'cd',
    'rad',
    # (Some) derived SI units
    'sr',
    'Hz',
    'N',
    'Pa',
    'J',
    'W',
    'C',
    'V',
    'F',
    'ohm',
    'S',
    'Wb',
    'T',
    'H',
    'lm',
    'lx',
    'Bq',
    'Gy',
    'Sv',
    'kat',
    # Some selected non-SI units, accepted for use with the SI
    'eV',           
    'l', 'L',       # liter
    't',            # ton
    'u',            # atomic mass unit
    'ua',           # astronomical unit
    # Some selected non-SI units, not accepted for use with SI (but included in this module)
    'torr',
    'Torr',
    'bar',
    'atm'
)

EMAX = 24

mul = {
    1: " ",
    3: "k",
    6: "M",
    9: "G",
    12: "T",
    15: "P",
    18: "E",
    21: "Z",
    24: "Y"
}

div = {
    1: " ",
    3: "m",
    6: "u",
    9: "n",
    12: "p",
    15: "f",
    18: "a",
    21: "z",
    24: "y"
}



#+---------------------------------------------------------+
#| Functions used to print values, uncertainties and units |
#+---------------------------------------------------------+

def print_exp(value, digits=0, align=False, debug=False):
    """ Returns a string expressing a number in exponential format 
        with a maximum number of significant digits in the mantissa

        This function returns a string representing a number in exponential format, with a maximum number 
        of significant digits in the mantissa
        If the given number is int, it will be converted to float
        Examples:
            print_exp(123456,    1) -> '1.0e5 '
            print_exp(-175836,   3) -> '-1.76e5'
            print_exp(0.123456,  4) -> '1.235e-1'
            print_exp(-0.912356, 1) -> '-9.0e-1'
            print_exp(312346149, 5) -> '3.1235e7'

        Parameters
        ----------

        value:  numerical value to be converted
        digits: maximum number of significant digits allowed for the mantissa
                0 or a negative value means keep all the digits
        align:  is set to True, add a leading space to integer number
                to mantain alignment to upper rows in columnar output

        Returns
        -------

        a string expressing the number in exponential format with the significant digits in the mantissa
    """

    value = float(value)

    # If the given number is nan, or inf, return the value as string
    if ( math.isnan(value) or math.isinf(value) ): return str(value)

    # If the number is negative apply the procedure to the positive part and then change sign
    if (value < 0):  return '-'+print_exp(-value, digits, debug)

    if (value == 0):  (mantissa, esp) = (0.0, 0)
    else:             (mantissa, esp) = _sep_mantissa_exp(value, digits=digits, debug=debug)

    ms=str(mantissa)
    es=str(esp)
    if (digits > 0): ms = ms.ljust(digits+1,'0')

    # If the mantissa string ends with '.0' cut the last two characters        
    if ( (ms[-1] == '0') and (ms[-2] == '.') ): ms = ms[0:-2]        

    # If the number is not decimal, add a leading space to preserve alignment
    if ( align and  ('.' not in ms)) : ms = ' ' + ms

    return ms + 'e' + es


def print_unit(value, unit, digits=0, kg_check=True, unit_check=False, align=False, debug=False):
    """ Returns a string expressing a physical quantity in SI (or compatible) units using prefixies.

        This function takes as arguments a numerical value and a string representig a measure unit
        It will convert the value using SI prefixs
        If the given number is of type int, it will be converted to float
        You can also specify how many significant digits must be provided in the result
        Examples:
            print_unit(1.23456E5,  'Hz')   -> '123.456 kHz'
            print_unit(1.23456E-2, 's', 3) -> '12.3 ms'
            print_unit(1.23456E-7, 'm', 3) -> '123.0 nm'

        WARNING: the function will (try to) treat nicely the case of 'kg', unless you disable this feature.
        However, you must give the unit correctly (e.g. you must provide 'kg' and *not* 'Kg')

        Parameters
        ----------

        value:          value of the physical quantity
        unit:           measure unit in which the physical quantity is expressed
        digits:         maximum number of digits to be shown (zero or negative number = no limit)
        kg_check:       if set to True, treat nicely the case of 'kg'
        unit_check:     if set to True, checks that the string given is included in the unit list
                        and returns an error message if it is not
        debug:          if set to True, print some additional information for debugging purposes

        Returns
        -------

        a string containing the same measure expressed with SI prefixs
    """

    if debug: print("*** function print_unit ***")    
    # If required, check that the given unit is included in the list
    if unit_check:
        if unit not in units: return('ERROR: \"'+unit+'\" is not a valid unit')

    # Try to treat nicely the ugly case of kg
    if kg_check & (unit == 'kg'):
        value = value * 1.0E3
        unit = 'g'

    # Do the conversion
    (v,u) = change_unit(value, unit, digits, debug)

    if debug:
        print("value  = ", v)
        print("unit   = ", u)

    # If nan or inf is returned
    if ( math.isnan(v) or math.isinf(v)):
        vs = str(v)
        us = str(u)
        if (digits > 0): vs = vs.rjust(digits+1)
        return vs + '  ' + us

    if (digits > 0):
        if (v > 0): vs = str(v).ljust(digits+1,'0')
        else:       vs = str(v).ljust(digits+2,'0')
    else:
        vs = str(v)
        
    us = str(u)

    if debug:
        print("vs     = ", vs)
        print("us     = ", us)

    # If the string value ends with '.0', and the final zero is not required, cut the last two characters
    if ( (vs[-1] == '0') and (vs[-2] == '.') and (len(vs) > digits+1) ): vs = vs[0:-2]

    # If the number is not decimal, add a leading space to preserve alignment
    if ( align and ('.' not in vs) ) : vs = ' ' + vs
        
    return vs + ' ' + us    


def print_uncertainty(value, uncertainty, digits=2, debug=False):
    """ Return a string expressing (value+/-uncertainty), properly fixing the number of significant digits.

        This function takes as arguments two numerical values representig a value and the associated uncertainty
        It will return a string representing the couple value +/- uncertainty, 
        fixing the numer of significant digits of the uncertainty as requested.

        Parameters
        ----------

        value:  numerical value to be converted
        digits: maximum number of significant digits allowed for the uncertainty
                0 or a negative value means keep all the digits
        align:  is set to True, add a leading space to integer number
                to mantain alignment to upper rows in columnar output

        Returns
        -------

        a string expressing the number in exponential format with the significant digits in the mantissa
    """

    (value, uncertainty, esp) = fix_uncertainty(value, uncertainty, digits, debug)
    vs = str(value)
    us = str(uncertainty)
    es = str(esp)
    # If the mantissa string ends with '.0' cut the last two characters        
    if ( (vs[-1] == '0') and (vs[-2] == '.') ): vs = vs[0:-2]
    if ( (us[-1] == '0') and (us[-2] == '.') ): us = us[0:-2] 
    # If the number is not decimal, add a leading space to preserve alignment
    if ('.' not in vs): vs = ' ' + vs
    if ('.' not in us): us = ' ' + us        
    string = '(' + vs + '+/-'+ us + ')'
    if (esp != 0): string = string +  'e' + es

    return string


#+----------------------------------------------------------+
#| Functions used to modify values, uncertainties and units |
#+----------------------------------------------------------+

def change_unit(value, unit, digits=0, debug=False):
    """ Express a physical quantity in SI (or compatible) units using prefixies.
                
        This function takes as arguments a numerical value and a string representing a measure unit
        It will convert the value using SI prefixes
        If the given number is int, it will be converted to float
        You can also specify how many significant digits must be provided in the result
        Examples:
            change_unit(1.23456E5,  'Hz')   -> (123.456, 'kHz')
            change_unit(1.23456E-2, 's', 3) -> (12.3, 'ms')
            change_unit(1.23456E-7, 'm', 3) -> (123.0, 'nm')

        WARNING: no check is made that the given string is a valid unit; 
        as a consequence the "ugly" case of "kg" is *not* treated nicely
        Examples:
            change_unit(1.23456E5,  'xy')   -> (0.123456, 'Mxy')

        Parameters
        ----------

        value:          value of the physical quantity
        unit:           measure unit in which the physical quantity is expressed
        digits:         maximum number of digits to be shown (zero or negative number = no limit)
        debug:          if set to True, print some additional information for debugging purposes

        Returns
        -------
        (new-value, new-unit)
        new-value:      the value of the same quantity expressed in the new unit
        new-unit:       the new unit, which is the old one multiplied by 10**(3*n)
                        where 3*n can be in range (-24, -21, -18, ..., 18, 21, 24)
    """

    if debug: print("*** function change_unit ***")
    
    value = float(value)

    # If the value is 0, inf, or nan, no conversion is needed
    if (math.isnan(value) or math.isinf(value) or (value==0.0)): return(value, unit)

    # If the value is negative, apply conversion to -value and than return -m
    if (value < 0):
        (m, prefix) = change_unit(-value, unit, digits=digits, debug=debug)
        return (-m, prefix)

    # Separate mantissa and exponent, taking care that mantissa is > 1
    (mantissa, esp) = _sep_mantissa_exp(value, digits=digits)

    if debug:
        print("mantissa = ", mantissa)
        print("exponent = ", esp)

    # Exponent is greater then the maximum value for which a prefix exist
    if (esp > EMAX):
        e = EMAX
        k = esp-e
        m = mantissa * 10**k
    # Exponent is lower then the minimum value for which a prefix exist
    elif (esp < -EMAX):
        e = -EMAX       
        k = esp-e
        m = mantissa * 10**k
    # Exponent is between minimum and maximum values: generate new mantissa and exponent
    else:
        k = esp % 3
        if (k==0):
            e = esp
            m = mantissa
        else:
            e = int(math.floor(esp / 3)) * 3
            m = mantissa * 10**k

    # If requested, reduce the number of decimal digits in the mantissa
    if (digits > 0):
        m = fix_digits(m, digits)
              
    if    (e > 0): prefix = mul[e]  + unit
    elif  (e < 0): prefix = div[-e] + unit
    else:          prefix = mul[1]  + unit

    # This is for debugging purposes
    if debug:
        print("m        = ", m)
        print("k        = ", k)
        print("e        = ", e)
        print("prefix   = ", prefix)

    if debug: print("*** end of function change_unit ***")

    return (m, prefix)
        

def fix_digits(value, digits, debug=False):
    """ Approximates a value to a given number of significant digits.
                
        This function approximate a real number to the number of significant digits specified
        If the given number is int, it will be converted to float
        Examples:
            fix_digits(1.23456,   1) ->  1.0
            fix_digits(-1.75836,  3) -> -1.76
            fix_digits(0.123456,  4) ->  0.1235
            fix_digits(-0.912356, 1) -> -1.0
            fix_digits(312346149, 5) ->  312350000.0

        Parameters
        ----------

        value:          numerical value to be converted
        digits:         number of signifincat digits required 
                        0 or a negative value means "do nothing at all"
        debug:          if set to True, print some additional information for debugging purposes

        Returns
        -------
                
        the nearest number with the requested significant digits
    """

    value = float(value)

    # If the number of digits requested is zero or less then zero, return the number without conversion
    if (digits <= 0):  return value
    # If the given number is zero, nan, or inf, do nothing at all
    if ( math.isnan(value) or math.isinf(value) or (value==0.0) ): return value
    # If the number is negative apply the procedure to the positive part and then change sign
    if (value < 0):  return -fix_digits(-value, digits, debug)
    # Apply the conversion
    (mantissa, esp) = _sep_mantissa_exp(value, digits=digits, debug=debug)

#    return mantissa*10**esp ###################################################################################
    return round(mantissa*10**esp, digits-esp)


def fix_uncertainty(value, uncertainty, digits=2, debug=False):
    """ Approximates a couple value to a given number of significant digits.
                
        Given a couple of real numbers, which represent the value and uncertainty of a physical quantity,
        this function approximate the uncertainty to the number of significant digits specified
        and the value to the  number of significant digits covered by the uncertainty.
        If the given number is int, it will be converted to float
        Examples:
            fix_uncertainty(1.23456,  0.034567, 1) ->   1.23,  0.03
            fix_uncertainty(-1.23456, 0.034567, 1) ->  -1.23,  0.03
            fix_uncertainty(-1.75836, 0.2345,   3) ->  -1.758, 0.234

        Parameters
        ----------

        value:          value of the physical quantity
        uncertainty:    uncertainty of the physical quantity, only the absolute value will be taken
        digits:         number of significant digits required for +the unceratinty 
                        0 or a negative value means "do nothing at all"
        debug:          if set to True, print some additional information for debugging purposes

        Returns
        -------
        the nearest number with the requested significant digits
    """

    value       = float(value)
    uncertainty = abs(float(uncertainty))
    digits      = int(digits)

    # If the number of digits requested is zero or less then zero, return the number without conversion
    if (digits <= 0):  return (value, uncertainty)
    # If the value or uncertaninty is zero, nan, or inf, do nothing at all
    if (math.isnan(value) or math.isinf(value) or math.isnan(uncertainty) or math.isinf(uncertainty)):
        return value, uncertainty
    # If the value is negative, apply the procedure to the positive part and then change sign
    if (value < 0):
        (v_mantissa, u_mantissa, v_esp) = fix_uncertainty(-value, uncertainty, digits, debug)
        return (-v_mantissa, u_mantissa, v_esp)
    # Fix the n
    (v_mantissa, v_esp) = _sep_mantissa_exp(value,       digits=None,   debug=debug)
    (u_mantissa, u_esp) = _sep_mantissa_exp(uncertainty, digits=digits, debug=debug)
    v_digits = v_esp - u_esp + digits
    if (v_mantissa >= 1): v_mantissa = round(v_mantissa, v_digits-1)
    else:                 v_mantissa = round(v_mantissa, v_digits)
    if (v_esp > u_esp):
        u_mantissa = u_mantissa * 10**(u_esp-v_esp)
        return (v_mantissa, u_mantissa, v_esp)
    else:
        v_mantissa = v_mantissa * 10**(v_esp-u_esp)
        return (v_mantissa, u_mantissa, u_esp)

        
def _sep_mantissa_exp(value, digits=None, debug=False):
    """ Separate mantissa and exponent of a floating point number, 
        limiting the number of significant digits in the mantissa.

        The function returns mantissa and exponent of a given number
        If required, also limits the number of the significant digits in the mantissa.

        Parameters
        ----------

        value:          numerical value to be converted
        digits:         maximum number of significant digits allowed for the mantissa
                        0 or a negative value means keep all the digits
        debug:          if set to True, print some additional information for debugging purposes

        Returns
        -------

        (mantissa, esp)
        mantissa: the mantissa (float)
        esp:      the exponent (int)
    """

    value = float(value)
        
    # If the given number is zero, nan, or inf, do nothing at all
    if ( math.isnan(value) or math.isinf(value) or (value==0.0) ): return value

    # Separate mantissa and exponent, taking care that mantissa is > 1
    if (value<0):
        esp      = int(math.floor(math.log10(-value)))
    else: 
        esp      = int(math.floor(math.log10(value)))
    mantissa = value / 10**esp

    # This is for debugging purposes
    if debug: print(mantissa, esp)

    # Reduce the number of decimal digits in the mantissa
    if (not( (digits is None) or math.isnan(digits) or math.isinf(digits) ) and (digits > 0)):
        digits = int(digits)
        if (abs(mantissa) >= 1):
            mantissa = round(mantissa, digits-1)
        else:
            mantissa = round(mantissa, digits)
        
    return (mantissa,esp)



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pysica",
    version="0.2.0",
    author="Pietro Mandracci",
    author_email="pietro.mandracci.software@gmail.com",
    description="PYthon tools for SImulation and CAlculus",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/pietromandracci/pysica",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires='>=3',
)



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Utility functions

 This subpackage contains some modules which provide several utility functions.

Modules
-------

*physics*
    functions used to calculate or convert some physical quantities
*optical*
    functions for analysis of the optical properties of thin films (absorption, reflection)
*mathematics*
    functions used to calculate averages, derivatives and other calculations
*pdf*
    collection of probability distribution functions used in the package
*statistics*
    some statistical tools
   
"""



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Functions for reading the particle properties from an ASCII file.

    This module contains functions for managing the import of particle properties from
    properly formatted ASCII files.

    Documentation is also available in the docstrings.
"""


# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

from pysica.parameters import *
from pysica.constants import *
from pysica.plasmapro.ccpla_defaults import *

# +--------------------------------------------------+
# | Read target particles properties from ascii file |
# +--------------------------------------------------+

def read_file_ntypes(filename, particles):
        """Gets the number of neutral types from the neutrals ascii file.

                Parameters
                ----------

                filename:  name of file containing neutral gases characteristics
                particles: an instance of the class TargetParticles defined in the module target_particles.py

                Returns
                -------

                (status, message)
                        status:  0 = no error
                                 1 = could not open file
                                 2 = no gas type given
                        message: a string containing an error message or 'Ok'
        """

        status, message  = 0, 'OK'              # This variable is used for error checking      

        try:
                f_input = open(filename,'r')
        except IOError:
                status  = 1
                message = 'could not open file'
                return (status, message)

        line            = 0     # Line of the ASCII file
        particles.types = 0
        for string in f_input:
                line = line + 1
 
                # Cut newline charater from the end of line
                string = string.strip(EOL)

                # Cut line portion at right of '#' char
                string = string.split('#')[0]

                # Cut spaces from beginning and end of line
                string = string.strip()

                # If the string is not empty, then the line defines a neutral type
                if (len(string)!=0):
                        particles.types= particles.types +1 

        f_input.close()

        if ( particles.types == 0 ):
                status  = 2
                message = 'no gas type was defined'

        return (status, message)



def read_file_properties(filename, sep, particles, debug=False):
        """Reads the neutrals characteristics from a ascii file and calculates some properties.

                        Parameters
                        ----------

                        filename:         name of file containing neutral gases characteristics
                        sep:              character used in the ASCII file to separate values (usually the tab char)
                        particles:        an instance of the class TargetParticles defined in the module target_particles.py

                        Initialized data attributes
                        ---------------------------

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = could not open file
                                         2 = missing values
                                         3 = numerical value expected
                                         4 = positive value expected
                                         5 = unknown molecule type
                                         6 = number of types exceedes max allowed
                                message: a string containing an error message or 'Ok'
                """

        status, message  = 0, 'OK'              # This variable is used for error checking      


        try:
                f_input = open(filename,'r')
        except IOError:
                status  = 1
                message = 'could not open file'
                return (status, message)

        line = 0        # Line of the ASCII file
        row  = 0        # Row of the file (excluding empty lines and comments)

        if debug: print('\nReading gas data from file :\"' + filename + '\"\n')

        for string in f_input:

                line = line + 1
                if debug: print("Line: " +str(line)+ " = \'" +string+ "\'")
 
                # Cut newline character from the end of line
                string = string.strip(EOL)

                # Cut line portion at right of '#' char
                string = string.split('#')[0]

                # Cut spaces from beginning and end of line
                string = string.strip()

                # Ignore the line if it is empty
                if (len(string)!=0):

                        # Get a list of the strings in the line
                        l = string.split(sep)
                        if (len(l) < 8):  # 8 is the number of strings expected for 1 excitation process and no dissociation
                                status  = 2
                                message = "line " + str(line) + ", missing values"
                                f_input.close()
                                return (status, message)
                        if debug: print("line, row, l = ", line, row, l)

                        # Get neutral name
                        particles.names[row] = l[0].strip()

                        # Get neutral molecule type
                        particles.molecule_type[row] = l[1].strip()
                        if (particles.molecule_type[row] not in MOLECULE_TYPES):        
                                status  = 5
                                message = ( "line "
                                            + str(line)
                                            + ", unknown molecule type "
                                            + "\""
                                            + str(particles.molecule_type[row])
                                            + "\" for "
                                            + str(particles.names[row]) )
                                f_input.close()
                                return (status, message)

                        # Get gas flow in arbitrary units
                        try:
                                value = float(l[2].strip())
                        except ValueError:
                                status  = 3
                                message = ( "line "
                                            + str(line)
                                            + ", numerical value expected for "
                                            + particles.names[row]
                                            + " gas flow" )
                                f_input.close()
                                return (status, message)
                        if (value <= 0):
                                status  = 4
                                message = ( "line "
                                            + str(line)
                                            + ", positive value expected for "
                                            + str(particles.names[row])
                                            + " gas flow" )
                                f_input.close()
                                return (status, message)                                
                        particles.gas_flow[row] = value

                        # Get neutral mass
                        try:
                                value = float(l[3].strip())
                        except ValueError:
                                status  = 3
                                message = "line "+str(line)+", numerical value expected for "+\
                                          particles.names[row]+" molecular mass"
                                f_input.close()
                                return (status, message)
                        if (value <= 0):
                                status  = 4
                                message = "line "+str(line)+", positive value expected for "+\
                                          particles.names[row]+" molecular mass"
                                f_input.close()
                                return (status, message)                        
                        particles.mass[row] = value

                        # Get secondary emission coefficient
                        try:
                                value = float(l[4].strip())
                        except ValueError:
                                status  = 3
                                message = "line "+str(line)+", numerical value expected for "+\
                                          particles.names[row]+" secondary emission coefficient"
                                f_input.close()
                                return (status, message)
                        if (value <= 0):
                                status  = 4
                                message = "line "+str(line)+", positive value expected for "+\
                                          particles.names[row]+" secondary emission coefficient"
                                f_input.close()
                                return (status, message)                                
                        particles.secondary_emission[row] = value

                        # Get first ionization energy
                        try:
                                value = float(l[5].strip())
                        except ValueError:
                                status  = 3
                                message = "line "+str(line)+", numerical value expected for "+\
                                          particles.names[row]+" ionization energy"
                                f_input.close()
                                return (status, message)
                        if (value <= 0):
                                status  = 4
                                message = "line "+str(line)+", positive value expected for "+\
                                          particles.names[row]+" ionization energy"
                                f_input.close()
                                return (status, message)                                
                        particles.ionization_energy[row] = value

                        # Get excitation energies
                        try:
                                value = int(l[6].strip())
                        except IndexError:
                                status  = 2
                                message = "line "+str(line)+", missing values"
                                f_input.close()
                                return (status, message)
                        except ValueError:
                                status  = 3
                                message = "line "+str(line) + \
                                          ", numerical value expected for number of excitation types "+\
                                          "of "+particles.names[row]
                                f_input.close()
                                return (status, message)
                        if (value < 1): 
                                status  = 4
                                message = "line "+str(line)+", positive value expected for number of excitation types "+\
                                          "of "+particles.names[row]
                                f_input.close()
                                return (status, message)
                        if (value > MAX_EXCITATION_TYPES): 
                                status  = 6
                                message = "line "+str(line)+", number of excitation types exceeds maximum allowed"+\
                                          " (" + str(MAX_EXCITATION_TYPES) + ')'
                                f_input.close()
                                return (status, message)                        
                        if (len(l) < value + 7): 
                                status  = 2
                                message = "line "+str(line)+", missing excitation energy values for "+\
                                          particles.names[row]
                                f_input.close()
                                return (status, message)
                        particles.excitation_types[row] = abs(value)
                        for i in range(value):
                                try:
                                        value2 = float(l[7+i].strip())
                                except ValueError:
                                        status  = 3
                                        message = "line " + str(line) + ", numerical value expected"+\
                                                  " for excitation energy of "+\
                                                  particles.names[row]
                                        f_input.close()
                                        return (status, message)
                                if (value2 <= 0):
                                        status  = 4
                                        message = "line " + str(line) + ", positive value expected"+\
                                                  " for excitation energy of "+\
                                                  particles.names[row]
                                        f_input.close()
                                        return (status, message)                                                
                                particles.excitation_energy[row,i] = abs(value2)
                        i_diss = 7 + value  # First index for dissociation data (if any)
                        if debug: print('value=', value, 'i_diss=', i_diss)

                        # If the molecule is not monoatomic, get dissociation energies
                        if (particles.molecule_type[row] != 'a'):
                                try:
                                        value = int(l[i_diss].strip())
                                except IndexError:
                                        status  = 2
                                        message = "line "+str(line)+", missing values"
                                        f_input.close()
                                        return (status, message)
                                except ValueError:
                                        status  = 3
                                        message = "line "+str(line) + \
                                                  ", numerical value expected for number of dissociation types "+\
                                                  "of "+particles.names[row]
                                        f_input.close()
                                        return (status, message)
                                if (value < 1): 
                                        status  = 4
                                        message = "line "+str(line)+", positive value expected for number of dissociation types "+\
                                                  "of "+particles.names[row]
                                        f_input.close()
                                        return (status, message)
                                if (value > MAX_DISSOCIATION_TYPES): 
                                        status  = 6
                                        message = "line "+str(line)+", number of dissociation types exceeds maximum allowed"+\
                                                  " (" + str(MAX_DISSOCIATION_TYPES) + ')'
                                        f_input.close()
                                        return (status, message)                                
                                if (len(l) < value + i_diss + 1): 
                                        status  = 2
                                        message = "line "+str(line)+", missing dissociation energy values for "+\
                                                  particles.names[row]
                                        f_input.close()
                                        return (status, message)
                                particles.dissociation_types[row] = abs(value)
                                for i in range(value):
                                        try:
                                                value2 = float(l[i_diss+i+1].strip())
                                        except ValueError:
                                                status  = 3
                                                message = "line " + str(line) + ", numerical value expected"+\
                                                          " for dissociation energy of "+\
                                                          particles.names[row]
                                                f_input.close()
                                                return (status, message)
                                        if (value2 <= 0):
                                                status  = 4
                                                message = "line " + str(line) + ", positive value expected"+\
                                                          " for dissociation energy of "+\
                                                          particles.names[row]
                                                f_input.close()
                                                return (status, message)                                                
                                        particles.dissociation_energy[row,i] = abs(value2)
                        row = row + 1

                if (row >= particles.types): break

        f_input.close()

        return (status, message)
